# SPDX-License-Identifier: BSD-2-Clause
# Copyright jdknight

# tar file attributes to check by default
DEFAULT_ATTRIBS = [
    'devmajor',
    'devminor',
    'gid',
    'gname',
    'linkname',
    'mode',
    'size',
    'type',
    'uid',
    'uname',
]

# other supported attributes to check for users requesting all attributes
EXTRA_ATTRIBS = [
    'chksum',
    'mtime',
    'offset',
    'offset_data',
]

# extended attributes that a user can use if explicitly opted for
EXTENDED_ATTRIBS = [
    'pax_headers',
]



# SPDX-License-Identifier: BSD-2-Clause
# Copyright jdknight

from pathlib import Path
from tardiff2.opts import TarDiffOpts
from tardiff2.tardiff import tardiff
from tests import TardiffTestCase
import tarfile


class TestDiffSameName(TardiffTestCase):
    def test_diff_same_name(self) -> None:
        sample1 = Path('container-1') / 'contents.txt'
        sample1.parent.mkdir()
        sample1.write_text('This is an example file.\n')
        self.assertTrue(sample1.is_file())

        sample2 = Path('container-2') / 'contents.txt'
        sample2.parent.mkdir()
        sample2.write_text('This is another example file.\n')
        self.assertTrue(sample2.is_file())

        example1 = Path('dist-1') / 'example.tgz'
        example1.parent.mkdir()
        with tarfile.open(example1, 'w') as tar:
            tar.add(sample1)
        self.assertTrue(example1.is_file())

        example2 = Path('dist-2') / 'example.tgz'
        example2.parent.mkdir()
        with tarfile.open(example2, 'w') as tar:
            tar.add(sample2)
        self.assertTrue(example2.is_file())

        files = [
            example1,
            example2,
        ]

        opts = TarDiffOpts()
        opts.detailed = 2

        diffed, _ = tardiff(files, opts=opts)
        self.assertTrue(diffed)



# SPDX-License-Identifier: BSD-2-Clause
# Copyright jdknight

from pathlib import Path
from tardiff2.opts import TarDiffOpts
from tardiff2.tardiff import tardiff
from tests import TardiffTestCase
import tarfile


class TestStrip(TardiffTestCase):
    def test_strip_two(self) -> None:
        sample1a = Path('container-1') / 'contents.txt'
        sample1a.parent.mkdir()
        sample1a.write_text('This is an example file.\n')
        self.assertTrue(sample1a.is_file())

        sample1b = Path('container-1') / 'nested-1' / 'contents.txt'
        sample1b.parent.mkdir()
        sample1b.write_text('This is an example file.\n')
        self.assertTrue(sample1b.is_file())

        sample2a = Path('container-2') / 'contents.txt'
        sample2a.parent.mkdir()
        sample2a.write_text('This is an example file.\n')
        self.assertTrue(sample2a.is_file())

        sample2b = Path('container-2') / 'nested-2' / 'contents.txt'
        sample2b.parent.mkdir()
        sample2b.write_text('This is an example file.\n')
        self.assertTrue(sample2b.is_file())

        example1 = Path('example1.tgz')
        with tarfile.open(example1, 'w') as tar:
            tar.add(sample1a)
            tar.add(sample1b)
        self.assertTrue(example1.is_file())

        example2 = Path('example2.tgz')
        with tarfile.open(example2, 'w') as tar:
            tar.add(sample2a)
            tar.add(sample2b)
        self.assertTrue(example2.is_file())

        files = [
            example1,
            example2,
        ]

        opts = TarDiffOpts()
        opts.strip = 2

        # should match the same even if we have different nested content
        # since we are stripping two levels
        diffed, _ = tardiff(files, opts=opts)
        self.assertFalse(diffed)

    def test_strip_zero(self) -> None:
        sample1 = Path('container-1') / 'contents.txt'
        sample1.parent.mkdir()
        sample1.write_text('This is an example file.\n')
        self.assertTrue(sample1.is_file())

        sample2 = Path('container-2') / 'contents.txt'
        sample2.parent.mkdir()
        sample2.write_text('This is an example file.\n')
        self.assertTrue(sample2.is_file())

        example1 = Path('example1.tgz')
        with tarfile.open(example1, 'w') as tar:
            tar.add(sample1)
        self.assertTrue(example1.is_file())

        example2 = Path('example2.tgz')
        with tarfile.open(example2, 'w') as tar:
            tar.add(sample2)
        self.assertTrue(example2.is_file())

        files = [
            example1,
            example2,
        ]

        opts = TarDiffOpts()
        opts.strip = 0

        # should be difference since we are not stripping and the base
        # container names are unique
        diffed, _ = tardiff(files, opts=opts)
        self.assertTrue(diffed)



# SPDX-License-Identifier: BSD-2-Clause
# Copyright jdknight

from __future__ import annotations
from pathlib import Path
from tempfile import TemporaryDirectory
import os
import unittest


class TardiffTestCase(unittest.TestCase):
    """
    a tardiff unit test case

    Provides a `unittest.TestCase` implementation that tardiff unit tests
    should inherit from. This test class provides the following capabilities:

    - Prepares a temporary working directory to allow creating test tar files
      into that will be automatically deleted.
    """

    def run(self, result: unittest.TestResult | None = None) -> None:
        """
        run the test

        Run the test, collecting the result into the TestResult object passed
        as result. See `unittest.TestCase.run()` for more details.

        Args:
            result (optional): the test result to populate
        """

        with TemporaryDirectory(prefix='.tardiff-') as tmpdir:
            owd = Path.cwd()
            try:
                os.chdir(tmpdir)
                super().run(result)
            finally:
                os.chdir(owd)



# SPDX-License-Identifier: BSD-2-Clause
# Copyright jdknight

from dataclasses import dataclass
from dataclasses import field
from tardiff2.defs import DEFAULT_ATTRIBS


@dataclass
class TarDiffOpts:
    attribs: list[str] = field(default_factory=lambda: DEFAULT_ATTRIBS)
    detailed: bool = False
    ignore_dirs: bool = False
    only_names: bool = False
    strip: int = 1



# SPDX-License-Identifier: BSD-2-Clause
# Copyright jdknight

from pathlib import Path
from tardiff2.tardiff import tardiff
from tests import TardiffTestCase
import platform
import tarfile


class TestMatchedLinkname(TardiffTestCase):
    def test_matched_linkname(self) -> None:
        if platform.system() != 'Linux':
            msg = 'symlink test skipped for non-Linux'
            raise self.skipTest(msg)

        sample1 = Path('container-1') / 'sample.txt'
        sample1.parent.mkdir()
        sample1.write_text('This is an example file.\n')
        self.assertTrue(sample1.is_file())

        sample1_link = Path('container-1') / 'sample-link'
        sample1_link.symlink_to('sample.txt')
        self.assertTrue(sample1_link.is_symlink())

        sample2 = Path('container-2') / 'sample.txt'
        sample2.parent.mkdir()
        sample2.write_text('This is an example file.\n')
        self.assertTrue(sample2.is_file())

        sample2_link = Path('container-2') / 'sample-link'
        sample2_link.symlink_to('sample.txt')
        self.assertTrue(sample2_link.is_symlink())

        example1 = Path('example1.tgz')
        with tarfile.open(example1, 'w') as tar:
            tar.add(sample1)
            tar.add(sample1_link)
        self.assertTrue(example1.is_file())

        example2 = Path('example2.tgz')
        with tarfile.open(example2, 'w') as tar:
            tar.add(sample2)
            tar.add(sample2_link)
        self.assertTrue(example2.is_file())

        files = [
            example1,
            example2,
        ]

        # should be difference with two different sized content
        diffed, _ = tardiff(files)
        self.assertFalse(diffed)



# SPDX-License-Identifier: BSD-2-Clause
# Copyright jdknight

from pathlib import Path
from tardiff2.tardiff import tardiff
from tests import TardiffTestCase
import tarfile


class TestDiffDevminor(TardiffTestCase):
    def test_diff_devminor(self) -> None:
        fentry = tarfile.TarInfo(name='container/dev-entry')
        fentry.type = tarfile.CHRTYPE
        fentry.devmajor = 1
        fentry.devminor = 3

        example1 = Path('example1.tgz')
        with tarfile.open(example1, 'w') as tar:
            tar.addfile(fentry)
        self.assertTrue(example1.is_file())

        fentry = tarfile.TarInfo(name='container/dev-entry')
        fentry.type = tarfile.CHRTYPE
        fentry.devmajor = 1
        fentry.devminor = 4

        example2 = Path('example2.tgz')
        with tarfile.open(example2, 'w') as tar:
            tar.addfile(fentry)
        self.assertTrue(example2.is_file())

        files = [
            example1,
            example2,
        ]

        # should be difference with two different devmodes
        diffed, _ = tardiff(files)
        self.assertTrue(diffed)



# SPDX-License-Identifier: BSD-2-Clause
# Copyright jdknight

from pathlib import Path
from tardiff2.tardiff import tardiff
from tests import TardiffTestCase
import shutil
import tarfile


class TestMatchedDuplicate(TardiffTestCase):
    def test_matched_duplicate(self) -> None:
        files = [
            'example.tar',
            'this-file-does-not-exist',
        ]

        sample = Path('container') / 'contents.txt'
        sample.parent.mkdir()
        sample.write_text('This is an example file.\n')
        self.assertTrue(sample.is_file())

        example1 = Path('example1.tgz')
        with tarfile.open(example1, 'w') as tar:
            tar.add(sample)
        self.assertTrue(example1.is_file())

        example2 = Path('example2.tgz')
        shutil.copyfile(example1, example2)
        self.assertTrue(example2.is_file())

        files = [
            example1,
            example2,
        ]

        # same tar file copied; should be a perfect match
        diffed, _ = tardiff(files)
        self.assertFalse(diffed)



# SPDX-License-Identifier: BSD-2-Clause
# Copyright jdknight

from pathlib import Path
from tardiff2.opts import TarDiffOpts
from tardiff2.tardiff import tardiff
from tests import TardiffTestCase
import tarfile


class TestMatchedOnlyNames(TardiffTestCase):
    def test_matched_only_names(self) -> None:
        sample1 = Path('container-1') / 'sample.txt'
        sample1.parent.mkdir()
        sample1.write_text('This is an example file.\n')
        self.assertTrue(sample1.is_file())

        sample2 = Path('container-2') / 'sample.txt'
        sample2.parent.mkdir()
        sample2.write_text('This is an example file.\n')
        self.assertTrue(sample2.is_file())

        def mode644(tarinfo: tarfile.TarInfo) -> tarfile.TarInfo:
            tarinfo.mode = 0o644
            return tarinfo

        example1 = Path('example1.tgz')
        with tarfile.open(example1, 'w') as tar:
            tar.add(sample1, filter=mode644)
        self.assertTrue(example1.is_file())

        def mode654(tarinfo: tarfile.TarInfo) -> tarfile.TarInfo:
            tarinfo.mode = 0o654
            return tarinfo

        example2 = Path('example2.tgz')
        with tarfile.open(example2, 'w') as tar:
            tar.add(sample2, filter=mode654)
        self.assertTrue(example2.is_file())

        files = [
            example1,
            example2,
        ]

        opts = TarDiffOpts()
        opts.only_names = True

        # while different modes, we are only comparing file names here
        diffed, _ = tardiff(files, opts=opts)
        self.assertFalse(diffed)



# SPDX-License-Identifier: BSD-2-Clause
# Copyright jdknight

from __future__ import annotations
from prettytable import PrettyTable
from tardiff2.opts import TarDiffOpts
from typing import Any
from typing import TYPE_CHECKING
import stat
import tarfile

if TYPE_CHECKING:
    from pathlib import Path


def tardiff(files: list[Path], *, opts: TarDiffOpts | None = None,
        quiet: bool = False) -> tuple[bool, dict[Path, Any]]:
    """
    perform a tar-diff

    Args:
        files: the list of tar files to compare
        opts (optional): the runtime options to consider
        quiet (optional): whether to dump results to the standard output stream

    Returns:
        tuple of the difference state and result data

    Raises:
        TardiffError
    """

    if not opts:
        opts = TarDiffOpts()

    if opts.strip < 0:
        msg = 'invalid strip-components value'
        raise TardiffInvalidStripCountError(msg)

    # unique file list
    files = list(dict.fromkeys(files))

    # should have at least two files
    if len(files) < 2:
        msg = 'need at least two files'
        raise TardiffRequireTwoFilesError(msg)

    # verify each file exists
    for file in files:
        if not file.is_file():
            msg = f'file is missing: {file}'
            raise TardiffFileMissingError(msg)

        if not tarfile.is_tarfile(file):
            msg = f'file is not a tar-file: {file}'
            raise TardiffFileInvalidError(msg)

    # compile a list of tar-file entries with tracked data
    entries: list[TarEntry] = []
    for file in files:
        with tarfile.open(file) as tar:
            entry = TarEntry(file)

            for member in tar.getmembers():
                parts = member.name.split('/', opts.strip)
                if len(parts) <= opts.strip:
                    continue

                new_name = parts[opts.strip]
                if member.isdir():
                    if opts.ignore_dirs:
                        continue
                    new_name += '/'

                entry.names.add(new_name)
                entry.mapping[new_name] = member

            entries.append(entry)

    entry0 = entries[0]

    # prepare diff results to capture
    has_differences = False
    mdb: dict[str, Any] = {}
    results: dict[Path, Any] = {}
    for entry in entries:
        results[entry.path] = {
            'extra': set(),
            'missing': set(),
            'modified': {},
        }

    # compile a list of files that exist in any of the provided archives as
    # well as a common set of files that exist in all provided archives
    all_names = entry0.names.copy()
    common_names = entry0.names.copy()
    for entry in entries[1:]:
        all_names.update(entry.names)
        common_names.intersection_update(entry.names)

    # track missing and extra files per file entry
    for entry in entries:
        missing_entries = all_names.difference(entry.names)
        extra_entries = entry.names.difference(common_names)

        results[entry.path]['missing'].update(missing_entries)
        results[entry.path]['extra'].update(extra_entries)

        if missing_entries or extra_entries:
            has_differences = True

    # look for modified files between each tarfile entry
    if not opts.only_names:
        for name in common_names:
            track: dict[Path, dict[str, str]] = {}
            base_tar_nfo = entry0.mapping[name]

            for entry in entries[1:]:
                node_tar_nfo = entry.mapping[name]

                # compare each attribute for differences
                for ac in opts.attribs:
                    base_val = getattr(base_tar_nfo, ac)
                    node_val = getattr(node_tar_nfo, ac)

                    if base_val != node_val:
                        if ac == 'mode':
                            base_val = _tardiff_modestr(base_val)
                            node_val = _tardiff_modestr(node_val)

                        track.setdefault(entry0.path, {})[ac] = base_val
                        track.setdefault(entry.path, {})[ac] = node_val

                # if we tracked any differences between this file as our base
                # entry, track in the modified database and our result set
                # the detected attribute differences
                if entry.path in track:
                    mdb.setdefault(name, {})[entry.path] = track[entry.path]
                    results[entry.path]['modified'][name] = track[entry.path]

            # if we tracked any differences, also include the base entry in
            # our modified database and result set
            if track:
                mdb.setdefault(name, {})[entry0.path] = track[entry0.path]
                results[entry0.path]['modified'][name] = track[entry0.path]
                has_differences = True

    # completed looking for differences; print results if configured for it
    if has_differences and not quiet:
        _tardiff_results(results, mdb, opts)

    # return the state to the caller
    return has_differences, results


def _tardiff_results(results: dict[Path, Any], modified_db: dict[str, Any],
        opts: TarDiffOpts) -> None:
    """
    print tardiff results to the standard output stream

    Args:
        results: the results of a tar-diff event
        modified_db: a more detailed modified database
        opts: the runtime options to consider
    """

    has_modified = False

    for entry, data in results.items():
        print()
        print(f'{entry.name}]')

        if data['extra']:
            print('  (extra entries)')
            for file in sorted(data['extra']):
                print(f'    {file}')

        if data['missing']:
            print('  (missing entries)')
            for file in sorted(data['missing']):
                print(f'    {file}')

        if data['modified']:
            has_modified = True

            if not opts.detailed:
                print('  (modified entries)')
                for file in sorted(data['modified']):
                    print(f'    {file}')

    if has_modified and opts.detailed:
        table = PrettyTable()

        # compile columns using file names
        field_names = ['Modified File']
        for mdb_entries in modified_db.values():
            field_names.extend(entry.name for entry in mdb_entries)
            break

        # unique column names
        seen_field_names: dict[str, int] = {}
        unique_field_names: list[str] = []

        for field_name in field_names:
            if field_name in seen_field_names:
                seen_field_names[field_name] += 1
                idx = seen_field_names[field_name]
                new_field_name = f'{field_name} ({idx})'
            else:
                seen_field_names[field_name] = 0
                new_field_name = field_name

            unique_field_names.append(new_field_name)

        # apply columns and other options
        table.field_names = unique_field_names
        table.align['Modified File'] = 'l'

        # build rows; the first column being the (stripped) file name entry
        # and trailing columns being the detected attributes that have a
        # difference compared to all the other files
        for fname, mdb_entries in sorted(modified_db.items()):
            values = [fname]
            values.extend(_attribstr(entry) for entry in mdb_entries.values())

            table.add_row(values, divider=True)

        print()
        print(table)


def _attribstr(value: dict[str, str]) -> str:
    return '\n'.join('{}={}'.format(*o) for o in value.items())


def _tardiff_modestr(value: int) -> str:
    """
    converts a tar member mode value into a mode string

    Args:
        value: the mode value

    Returns:
        the mode string
    """
    return stat.filemode(int(oct(value), 8))[1:]


class TarEntry:
    def __init__(self, path: Path) -> None:
        """
        tardiff tar entry

        Holds a series of tracked results for a given tarfile tracked in
        a tardiff invoke.

        Args:
            path: the path of the file

        Attributes:
            mapping: a map from a file name entry to a cached TarInfo instance
            names: the set of detected filenames in the tarfile
            path: the path of the file
        """
        self.mapping: dict[str, tarfile.TarInfo] = {}
        self.names: set[str] = set()
        self.path = path


class TardiffError(Exception):
    pass


class TardiffFileInvalidError(TardiffError):
    pass


class TardiffFileMissingError(TardiffError):
    pass


class TardiffInvalidStripCountError(TardiffError):
    pass


class TardiffRequireTwoFilesError(TardiffError):
    pass



# SPDX-License-Identifier: BSD-2-Clause
# Copyright jdknight

from pathlib import Path
from tardiff2 import __version__ as tardiff2_version
from tardiff2.defs import DEFAULT_ATTRIBS
from tardiff2.defs import EXTENDED_ATTRIBS
from tardiff2.defs import EXTRA_ATTRIBS
from tardiff2.opts import TarDiffOpts
from tardiff2.tardiff import tardiff
import argparse
import itertools
import os


def main() -> None:
    """
    mainline

    The mainline for tardiff.
    """

    try:
        parser = argparse.ArgumentParser(prog='tardiff2')
        parser.add_argument('file', nargs='+',
            help='Files to compare against')
        parser.add_argument('--all', action='store_true',
            help='Compare all known attributes')
        parser.add_argument('--ignore-directories', '-D', action='store_true',
            help='Ignore directory entries')
        parser.add_argument('--more', '-m', action='store_true',
            help='Show a more detailed comparison of file differences')
        parser.add_argument('--names', '-n', action='store_true',
            help='Only compare file names and not attributes')
        parser.add_argument('--strip-components', type=int, default=1,
            help='Strip given number of leading components from file names')
        parser.add_argument('--version', action='version',
            version='%(prog)s ' + tardiff2_version)
        parser.add_argument('--[no-]ATTRIB', action='store_true',
            help='Attributes to check or ignore')

        for entry in itertools.chain(
                DEFAULT_ATTRIBS, EXTRA_ATTRIBS, EXTENDED_ATTRIBS):
            parser.add_argument(f'--{entry}', action='store_true',
                help=argparse.SUPPRESS)
            parser.add_argument(f'--no-{entry}', action='store_true',
                help=argparse.SUPPRESS)

        args = parser.parse_args()

        files = []
        for fentry in args.file:
            path = Path(os.fsdecode(fentry))
            files.append(path)

        # compile a list of attributes to use
        attribs = set(DEFAULT_ATTRIBS)

        if args.all:
            attribs.update(EXTRA_ATTRIBS)

        for entry in itertools.chain(
                DEFAULT_ATTRIBS, EXTRA_ATTRIBS, EXTENDED_ATTRIBS):
            if getattr(args, entry):
                attribs.add(entry)

        for entry in itertools.chain(
                DEFAULT_ATTRIBS, EXTRA_ATTRIBS, EXTENDED_ATTRIBS):
            if getattr(args, f'no_{entry}'):
                attribs.remove(entry)

        # compile options to use for the tardiff check
        opts = TarDiffOpts(
            attribs=sorted(attribs),
            detailed=args.more,
            ignore_dirs=args.ignore_directories,
            only_names=args.names,
            strip=args.strip_components,
        )

        # perform a difference check of all provided tar files
        has_differences, _ = tardiff(files, opts=opts)
        if has_differences:
            raise SystemExit(1)

    except KeyboardInterrupt:
        print()


if __name__ == '__main__':
    main()



# SPDX-License-Identifier: BSD-2-Clause
# Copyright jdknight

from pathlib import Path
from tardiff2.opts import TarDiffOpts
from tardiff2.tardiff import TardiffFileInvalidError
from tardiff2.tardiff import TardiffFileMissingError
from tardiff2.tardiff import TardiffInvalidStripCountError
from tardiff2.tardiff import TardiffRequireTwoFilesError
from tardiff2.tardiff import tardiff
import unittest


class TestOpts(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        tests_dir = Path(__file__).parent.resolve()
        cls.assets_dir = tests_dir / 'assets'

    def test_opts_default_strip_one(self) -> None:
        opts = TarDiffOpts()
        self.assertEqual(opts.strip, 1)

    def test_opts_tardiff_invalid_strip(self) -> None:
        opts = TarDiffOpts()
        opts.strip = -2

        with self.assertRaises(TardiffInvalidStripCountError):
            tardiff([], opts=opts)

    def test_opts_tardiff_missing(self) -> None:
        files = [
            self.assets_dir / 'example.tar',
            self.assets_dir / 'this-file-does-not-exist',
        ]

        with self.assertRaises(TardiffFileMissingError):
            tardiff(files)

    def test_opts_tardiff_non_tarfile(self) -> None:
        files = [
            self.assets_dir / 'example.tar',
            self.assets_dir / 'non-tar-file.txt',
        ]

        with self.assertRaises(TardiffFileInvalidError):
            tardiff(files)

    def test_opts_tardiff_not_enough_files(self) -> None:
        files = []
        with self.assertRaises(TardiffRequireTwoFilesError):
            tardiff(files)

        files = [
            self.assets_dir / 'example.tar',
        ]
        with self.assertRaises(TardiffRequireTwoFilesError):
            tardiff(files)

        files = [
            self.assets_dir / 'example.tar',
            self.assets_dir / 'example.tar',
        ]
        with self.assertRaises(TardiffRequireTwoFilesError):
            tardiff(files)
